"""
Implementation of the LogBook class.

LogBook class provides an interface to persist the logs on the filesystem,
tensorboard, remote backends, etc.

"""

import importlib
import time
from copy import deepcopy
from typing import List, Optional

from ml_logger.logger.base import Logger as LoggerType
from ml_logger.types import ConfigType, KeyMapType, LogType, MetricType


class LogBook:
    """This class provides an interface to persist the logs on the filesystem,
    tensorboard, remote backends, etc.

    """

    def __init__(self, config: ConfigType):
        """Initialise the Logbook class

        Args:
            logbook_config (ConfigType): Config to initialise the
                LogBook class. The logbook config must have the
                following keys:
                id: Id of the current LogBook instance. This
                    attribute is logged with each log and is useful
                    when multiple LogBook instances are needed (for
                    example with multiprocessing)
                logger_file_path: Path to the file, where the logs
                    will be written
                The logbook config can be created using the make_config
                method defined in ml_logger/logbook.py
            config (ConfigType): config corresponding to the ml experiment
                creating the logbook
        """
        self.id = config["id"]
        self.logger_name = config["name"]
        self.time_format = "%I:%M:%S%p %Z %b %d, %Y"  # 10:21:14PM EST Mar 04, 2020
        self.loggers: List[LoggerType] = []
        for logger_name, logger_config in config["loggers"].items():
            logger_module = importlib.import_module(f"ml_logger.logger.{logger_name}")
            logger_cls = getattr(logger_module, "Logger")
            logger = logger_cls(config=logger_config)
            self.loggers.append(logger)

    def _process_log(self, log: LogType, log_type: str) -> LogType:
        """Process the log before writing

        Args:
            log (LogType): Log to process
            log_type (str): Type of the log: config, metric, metadata, etc

        Returns:
            LogType: Processed log
        """
        log["logbook_id"] = self.id
        log["logbook_timestamp"] = time.strftime(self.time_format)
        log["logbook_type"] = log_type
        return log

    def write_log(self, log: LogType, log_type: str = "metric") -> None:
        """Write log to loggers

        Args:
            log (LogType): Log to write
            log_type (str, optional): Type of this log. Defaults to "metric".
        """

        log = self._process_log(log, log_type)
        for logger in self.loggers:
            logger.write_log(log=deepcopy(log))

    def write_config_log(self, config: ConfigType) -> None:
        """Write config to loggers

        Args:
            config [ConfigType]: Config to write.
        """
        return self.write_log(log=config, log_type="config")

    def write_metric_log(self, metric: MetricType) -> None:
        """Write metric to loggers

        Args:
            metric (MetricType): Metric to write
        """
        return self.write_log(log=metric, log_type="metric")

    def write_message(self, message: str, log_type: str = "info") -> None:
        """Write message string to loggers

        Args:
            message (str): Message string to write
            log_type (str, optional): Type of this message (log).
                Defaults to "info".
        """
        return self.write_log(log={"messgae": message}, log_type=log_type)

    def write_metadata_log(self, metadata: LogType) -> None:
        """Write metadata to loggers

        Args:
            metadata (LogType): Metadata to wite
        """
        return self.write_log(log=metadata, log_type="metadata")


def make_config(
    id: str = "0",
    name: str = "default_logger",
    write_to_console: bool = True,
    logger_dir: Optional[str] = None,
    filename: Optional[str] = None,
    filename_prefix: str = "",
    create_multiple_log_files: bool = True,
    wandb_config: Optional[ConfigType] = None,
    wandb_key_map: Optional[KeyMapType] = None,
    wandb_prefix_key: Optional[str] = None,
    tensorboard_config: Optional[ConfigType] = None,
    tensorboard_key_map: Optional[KeyMapType] = None,
    tensorboard_prefix_key: Optional[str] = None,
    mlflow_config: Optional[ConfigType] = None,
    mlflow_key_map: Optional[KeyMapType] = None,
    mlflow_prefix_key: Optional[str] = None,
) -> ConfigType:

    """Make the config that can be passed to the LogBook constructor

    Args:
        id (str, optional): Id of the current LogBook instance. Defaults to "0".
        name (str, optional): Name of the logger. Defaults to "default_logger".
        write_to_console (bool, optional): Should write the logs to console.
            Defaults to True
        logger_dir (str, optional):  Path where the logs will be
            written. If None is passed, logs are not written to the filesystem.
            LogBook creates the directory, if it does not exist. Defaults
            to None.
        filename (str, optional):  Name to assign to the log file (eg
            log.jsonl). If None is passed, this argument is ignored. If
            the value is set, `filename_prefix` and `create_multiple_log_files`
            arguments are ignored. Defaults to None.
        filename_prefix (str): String to prefix before the name of the log
            files. Eg if filename_prefix is "dummy", name of log files are
            dummymetric.jsonl, dummylog.jsonl etc. This argument is ignored
            if `filename` is set. Defaults to "".
        create_multiple_log_files (bool, optional): Should multiple log
            files be created - for config, metric, metadata and message
            logs. If True, the files are named as config_log.jsonl,
            metric_log.jsonl etc. If False, only one file log.jsonl is
            created. This argument is ignored if `filename` is set.
            Defaults to True.
        wandb_config (Optional[ConfigType], optional): Config for the wandb
            logger. If None, wandb logger is not created. The config can
            have any parameters that wandb.init() methods accepts
            (https://docs.wandb.com/library/init). Note that the wandb_config
            is passed as keyword arguments to the wandb.init() method.
            This provides a lot of flexibility to the users to configure
            wandb. This also means that the config should not have any
            parameters that wandb.init() would not accept. Defaults to None.
        wandb_key_map (Optional[KeyMapType], optional): When using wandb
            logger for logging metrics, certain keys are required. This
            dictionary provides an easy way to map the keys in the `log`
            to be written) with the keys that wandb logger needs. For
            instance, wandb logger needs a `step` key in all the metric
            logs. If your logs have a key called `epoch` that you want to
            use as `step`, set `wandb_key_map` as `{epoch: step}`. This
            argument is ignored if set to None. Defaults to None.
        wandb_prefix_key (Optional[str], optional): When a metric is logged
            to wandb, prefix the value (corresponding to the key) to all
            the remaining keys before values are logged in the wandb logger.
            This argument is ignored if set to None. Defaults to None.
        tensorboard_config (Optional[ConfigType], optional): config to
            initialise the tensorboardX logger. The config can have
            any parameters that tensorboardX.SummaryWriter() method accepts
            (https://tensorboardx.readthedocs.io/en/latest/tensorboard.html#tensorboardX.SummaryWriter).
            Note that the config is passed as keyword arguments to the
            tensorboardX.SummaryWriter() method. This provides a lot
            of flexibility to the users to configure tensorboard. This also
            means that config should not have any parameters that
            tensorboardX.SummaryWriter() would not accept. Defaults to None.
        tensorboard_key_map (Optional[KeyMapType], optional): When using
            tensorboard logger for logging metrics, certain keys are required.
            This dictionary provides an easy way to map the keys in the `log`
            (to be written) with the keys that tensorboard logger needs.
            For instance, tensorboard logger needs a `main_tag` key and a
            `global_step` in all the metric logs. If your logs have a key
            called `epoch` that you want to use as `step`, and a key called
            `mode` that you want to use as `main_tag`, set `tensorboard_key_map`
            as `{epoch: global_step, mode: main_tag}`. This argument is
             ignored if set to None. Defaults to None.
        tensorboard_prefix_key (Optional[str], optional): When a metric is
            logged to tensorboard, prefix the value (corresponding to the key)
            to all the remaining keys before values are logged in the
            tensorboard logger. This argument is ignored if set to None.
            Defaults to None.
        mlflow_config (Optional[ConfigType], optional): config to
            initialise an mlflow experiment. The config can have
            any parameters that mlflow.create_experiment() method accepts
            (https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.create_experiment).
            Note that the config is passed as keyword arguments to the
            mlflow.create_experiment() method. This provides a lot
            of flexibility to the users to configure mlflow. This also
            means that config should not have any parameters that
            mlflow.create_experiment would not accept. Defaults to None.
        mlflow_key_map (Optional[KeyMapType], optional): When using mlflow
            logger for logging metrics, certain keys are required. This
            dictionary provides an easy way to map the keys in the `log`
            (to be written) with the keys that mlflow logger needs. For
            instance, mlflow logger needs a `step` key in all the metric
            logs. If your logs have a key called `epoch` that you want to
            use as `step`, set `mlflow_key_map` as `{epoch: step}`. This
            argument is ignored if set to None. Defaults to None.
        mlflow_prefix_key (Optional[str], optional): When a metric is logged
            to mlflow, prefix the value (corresponding to the key) to all
            the remaining keys before values are logged in the mlflow logger.
            This argument is ignored if set to None. Defaults to None.
    Returns:
        ConfigType: config to construct the LogBook
    """

    loggers: ConfigType = {}
    if logger_dir is not None:
        loggers["filesystem"] = {
            "logger_dir": logger_dir,
            "logger_name": name,
            "write_to_console": write_to_console,
            "filename": filename,
            "create_multiple_log_files": create_multiple_log_files,
            "filename_prefix": filename_prefix,
        }
        loggers["filesystem"]["logbook_key_map"] = None
        loggers["filesystem"]["logbook_key_prefix"] = None
    if wandb_config is not None:
        loggers["wandb"] = wandb_config
        loggers["wandb"]["logbook_key_map"] = wandb_key_map
        loggers["wandb"]["logbook_key_prefix"] = wandb_prefix_key

    if tensorboard_config is not None:
        loggers["tensorboard"] = tensorboard_config
        loggers["tensorboard"]["logbook_key_map"] = tensorboard_key_map
        loggers["tensorboard"]["logbook_key_prefix"] = tensorboard_prefix_key

    if mlflow_config is not None:
        loggers["mlflow"] = mlflow_config
        loggers["mlflow"]["logbook_key_map"] = mlflow_key_map
        loggers["mlflow"]["logbook_key_prefix"] = mlflow_prefix_key

    config = {"id": id, "name": name, "loggers": loggers}
    return config



import pathlib
from typing import Any, Dict, List, Tuple


def flatten_dict(
    d: Dict[str, Any], parent_key: str = "", sep: str = "#"
) -> Dict[str, Any]:
    """Flatten a given dict using the given seperator.

    Taken from https://stackoverflow.com/a/6027615/1353861

    Args:
        d (Dict[str, Any]): dictionary to flatten
        parent_key (str, optional): Keep track of the higher level key
            Defaults to "".
        sep (str, optional): string for concatenating the keys. Defaults
            to "#"

    Returns:
        Dict[str, Any]: [description]
    """
    items: List[Tuple[str, Any]] = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def make_dir(path: str) -> None:
    """Make dir, if not exists

    Args:
        path (str): dir to make
    """
    pathlib.Path(path).mkdir(parents=True, exist_ok=True)



"""Abstract logger class"""
from abc import ABCMeta, abstractmethod
from typing import List, Optional

from ml_logger.types import ConfigType, KeyMapType, LogType


class Logger(metaclass=ABCMeta):
    """Abstract Logger Class
    """

    @abstractmethod
    def __init__(self, config: ConfigType):
        """Initialise the Logger

        Args:
            config (ConfigType): config to initialise the logger
        """
        self.keys_to_retain: Optional[List[str]] = None
        self.keys_to_skip: Optional[List[str]] = None
        self.keys_to_check: Optional[List[str]] = None
        self.key_map: Optional[KeyMapType] = config.pop("logbook_key_map")
        self.key_prefix: Optional[str] = config.pop("logbook_key_prefix")

    @abstractmethod
    def write_log(self, log: LogType) -> None:
        """Interface to write the log

        Args:
            log (LogType): Log to write
        """
        pass

    def _prepare_log_to_write(self, log: LogType) -> LogType:
        """Remove certain keys before writing the log.

        LogBook adds some keys to track metadata. These keys are filtered
        for some loggers (like wandb).

        `self.keys_to_retain` informs what keys are to be retained. All
        keys are retained if `self.keys_to_retain` is None.

        `self.keys_to_skip` informs what keys are to be skipped. No key
        is skipped if `self.keys_to_skip` is None.

        Args:
            log (LogType): Log to write

        Returns:
            LogType: Log with certain keys removed
        """
        processed_log: LogType = {}

        if self.keys_to_retain is None:
            # Retain all the keys, other than the ones to skip
            processed_log = log
        else:
            # Retain only the keys that need to be retained
            for key in self.keys_to_retain:
                processed_log[key] = log[key]

        if self.keys_to_skip is not None:
            for key in self.keys_to_skip:
                if key in processed_log:
                    processed_log.pop(key)
        return processed_log

    def _prepare_metric_log_to_write(self, log: LogType) -> LogType:
        """Map some keys to another keys, remove some keys before writing
        the log.

        Some loggers require specific keys to be present. User can specify
        those keys via a different name and should be mapped to the required
        name before passing to the logger.

        LogBook adds some keys to track metadata. These keys are filtered
        for some loggers (like wandb).

        `self.keys_to_retain` informs what keys are to be retained. All
        keys are retained if `self.keys_to_retain` is None.

        `self.keys_to_skip` informs what keys are to be skipped. No key
        is skipped if `self.keys_to_skip` is None.

        Args:
            log (LogType): Log to write

        Returns:
            LogType: Log with certain keys removed
        """

        if self.key_map is not None:
            for key, mapped_key in self.key_map.items():
                log[mapped_key] = log.pop(key)
        return self._prepare_log_to_write(log=log)



"""Implementation of Parser to parse the logs"""

import json
from collections.abc import Sequence
from functools import partial
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union, overload

import pandas as pd

from ml_logger.logger import filesystem
from ml_logger.parser import parser as base_parser
from ml_logger.parser import utils as parser_utils
from ml_logger.parser.config import Parser as ConfigParser
from ml_logger.parser.metric import Parser as MetricParser
from ml_logger.types import LogType


class Experiment:
    def __init__(
        self, config: LogType, metrics: Dict[str, pd.DataFrame], logger_dir: str,
    ):
        self.config = config
        self.metrics = metrics
        self.logger_dir = logger_dir


class ExperimentSequence(Sequence):  # type: ignore
    """Provides a list-like interface to a collection of Experimente
    """

    def __init__(self, experiments: List[Experiment]):
        self.experiments = experiments
        super().__init__()

    def groupby(
        self, group_fn: Callable[[Experiment], str]
    ) -> Dict[str, "ExperimentSequence"]:
        """Group experiments in the sequence

        Args:
            group_fn: Function to assign a string group id to the experiment

        Returns:
            Dict[str, ExperimentSequence]: A dictionary mapping the sring
            group id to a sequence of experiments
        """
        grouped_experiments: Dict[str, List[Experiment]] = {}
        for experiment in self.experiments:
            key = group_fn(experiment)
            if key not in grouped_experiments:
                grouped_experiments[key] = []
            grouped_experiments[key].append(experiment)

        return {
            key: ExperimentSequence(value) for key, value in grouped_experiments.items()
        }

    def filter(self, filter_fn: Callable[[Experiment], bool]) -> "ExperimentSequence":
        """Filter experiments in the sequence

        Args:
            filter_fn: Function to filter an experiment

        Returns:
            ExperimentSequence: A sequence of experiments for which the
            filter condition is true
        """
        return ExperimentSequence(
            [experiment for experiment in self.experiments if filter_fn(experiment)]
        )

    @overload  # noqa:F811
    def __getitem__(self, index: int) -> Experiment:
        pass

    @overload  # noqa:F811
    def __getitem__(self, index: slice) -> "ExperimentSequence":
        pass

    def __getitem__(  # noqa:F811
        self, index: Union[int, slice]
    ) -> Union[Experiment, "ExperimentSequence"]:
        if isinstance(index, slice):
            return ExperimentSequence(self.experiments[index])
        return self.experiments[index]

    def __len__(self) -> int:
        return len(self.experiments)


class Parser(base_parser.Parser):
    """Class to parse the entire experiment from the log dir
    """

    def __init__(
        self,
        log_transformer: Callable[
            [LogType], LogType
        ] = parser_utils.identity_log_transformer,
        error_handler: Callable[
            [str, json.decoder.JSONDecodeError], Optional[LogType]
        ] = parser_utils.silent_error_handler,
    ):
        """Class to parse the experiment from the log files

        Args:
            log_transformer (Callable[[LogType], LogType], optional):
                Function to transform the logs after reading them from the
                filesystem. Defaults to parser_utils.identity_log_transformer.
            error_handler (Callable[[str, json.decoder.JSONDecodeError],
                Optional[LogType]], optional): Function to handle the
                error when the parser reads an invalid json string.
                Defaults to parser_utils.silent_error_handler.
        """

        super().__init__(
            log_transformer=log_transformer, error_handler=error_handler,
        )
        self.config_parser = ConfigParser(
            log_transformer=log_transformer, error_handler=error_handler
        )

        self.metric_parser = MetricParser(
            log_transformer=log_transformer, error_handler=error_handler
        )
        self.log_type = None

    def get_one_experiment(
        self, logger_dir: str, filename: Optional[str] = None, filename_prefix: str = ""
    ) -> Experiment:
        """Load one experiment from the log dir

        Args:
            logger_dir (str): Path where the logs will be read from.
            filename (Optional[str], optional): Name to assigned to the
            log file (eg log.jsonl). If None is passed, this argument is
            ignored and it is assumed that multiple log files are written
            per experiment (for config, metadata, metric etc).  If the
            value is set, `filename_prefix` argument is ignored. Defaults
            to None.
            filename_prefix (str, optional): String to prefix before the
            name of the log files. Eg if filename_prefix is "dummy",
            name of log files are dummymetric.jsonl, dummylog.jsonl etc.
            This argument is ignored if `filename` is set. Defaults to "".

        Returns:
            Experiment
        """

        if filename is None:
            # We need to parse multiple files

            get_file_path = partial(
                filesystem.get_logger_file_path,
                logger_dir=logger_dir,
                filename=filename,
                filename_prefix=filename_prefix,
            )

            config_file_path = get_file_path(filename_suffix="config_log")
            metric_file_path = get_file_path(filename_suffix="metric_log")

            config = self.config_parser.get_one_config(file_path=config_file_path)
            metrics = self.metric_parser.get_metrics_as_df(file_path=metric_file_path)

        else:
            # We need to parse just one file
            get_file_path = partial(
                filesystem.get_logger_file_path,
                logger_dir=logger_dir,
                filename=filename,
                filename_prefix=filename_prefix,
            )

            log_file_path = get_file_path(filename_suffix="log",)

            config = self.config_parser.get_one_config(file_path=log_file_path)
            metrics = self.metric_parser.get_metrics_as_df(file_path=log_file_path)

        assert config is not None
        return Experiment(config=config, metrics=metrics, logger_dir=logger_dir)

    def get_experiment_sequence(
        self,
        logger_path: str,
        filename: Optional[str] = None,
        filename_prefix: str = "",
    ) -> ExperimentSequence:
        """Load a sequence of experiments frrom the log dirs.

        Args:
            logger_dir (str): Path where the logs will be read from.
            filename (Optional[str], optional): Name to assigned to the
            log file (eg log.jsonl). If None is passed, this argument is
            ignored and it is assumed that multiple log files are written
            per experiment (for config, metadata, metric etc).  If the
            value is set, `filename_prefix` argument is ignored. Defaults
            to None.
            filename_prefix (str, optional): String to prefix before the
            name of the log files. Eg if filename_prefix is "dummy",
            name of log files are dummymetric.jsonl, dummylog.jsonl etc.
            This argument is ignored if `filename` is set. Defaults to "".

        Returns:
            ExperimentSequence
        """

        experiments = []
        for current_dir in Path(logger_path).iterdir():
            if current_dir.is_dir():
                experiments.append(
                    self.get_one_experiment(
                        logger_dir=str(current_dir),
                        filename=filename,
                        filename_prefix=filename_prefix,
                    )
                )
        return ExperimentSequence(experiments=experiments)



"""Implementation of Parser to parse the logs"""

import json
from typing import Callable, Iterator, Optional

from ml_logger.parser import utils as parser_utils
from ml_logger.types import LogType


class Parser:
    """Class to parse the log files
    """

    def __init__(
        self,
        log_transformer: Callable[
            [LogType], LogType
        ] = parser_utils.identity_log_transformer,
        error_handler: Callable[
            [str, json.decoder.JSONDecodeError], Optional[LogType]
        ] = parser_utils.silent_error_handler,
    ):
        """Class to parse the log files

        Args:
            log_transformer (Callable[[LogType], LogType], optional):
                Function to transform the logs after reading them from the
                filesystem. Defaults to parser_utils.identity_log_transformer.
            error_handler (Callable[[str, json.decoder.JSONDecodeError],
                Optional[LogType]], optional): Function to handle the
                error when the parser reads an invalid json string.
                Defaults to parser_utils.silent_error_handler.
        """
        self.log_transformer = log_transformer
        self.error_handler = error_handler
        self.log_type: Optional[str] = None

    def filter_log(self, log: LogType) -> bool:
        """Check if the log is a valid log

        Args:
            log (LogType): log to check

        Returns:
            bool: True if the log is a valid log
        """
        if self.log_type is None:
            return True
        for key in ["logbook_type", "type"]:
            if key in log and log[key] == self.log_type:
                return True

        return False

    def _parse_file(self, file_path: str) -> Iterator[LogType]:
        """Method to open a log file and parse the logs

        Args:
            file_path (str): Log file to read from

        Returns:
            Iterator[LogType]: Iterator over the logs

        Yields:
            Iterator[LogType]: Iterator over the logs
        """
        with open(file_path) as f:
            for line in f:
                try:
                    yield json.loads(line)
                except json.decoder.JSONDecodeError as e:
                    # This could be the scase where a new line was missing
                    error_handler_response = self.error_handler(line, e)
                    if error_handler_response is not None:
                        yield error_handler_response

    def get_logs(self, file_path: str) -> Iterator[LogType]:
        """Method to open a log file, parse the logs and return logs

        Args:
            file_path (str): Log file to read from

        Returns:
            Iterator[LogType]: Iterator over the logs

        Yields:
            Iterator[LogType]: Iterator over the logs
        """
        for log in self._parse_file(file_path=file_path):
            if self.filter_log(log):
                yield self.log_transformer(log)



"""Implementation of different type of metrics"""

import operator
from typing import Any, Iterable, Optional, Union

from ml_logger.types import ComparisonOpType, LogType, NumType, ValueType


class BaseMetric:
    """Base Metric class. This class is not to be used directly

    """

    def __init__(self, name: str):
        """Base Metric class

        All the metrics extend this class. It is not to be used directly

        Args:
            name (str): Name of the metric
        """
        self.name = name
        self.val: ValueType
        self.reset()

    def reset(self) -> None:
        """Reset the metric to the default value
        """
        self.val = 0

    def update(self, val: Any) -> None:
        """Update the metric using the current val

        Args:
            val (Any): Current value. This value is used to update the
                metric
        """
        pass

    def get_val(self) -> ValueType:
        """Get the current value of the metric

        """
        return self.val

    def __str__(self) -> str:
        return str(self.get_val())

    def __repr__(self) -> str:
        return f"{self.__class__} {self.__dict__}"


class CurrentMetric(BaseMetric):
    """Metric to track only the most recent value

    Args:
        BaseMetric: Base metric class
    """

    def __init__(self, name: str):
        super().__init__(name)

    def update(self, val: ValueType) -> None:
        """Update the metric using the current val

        Args:
            val (Any): Current value. The metric value is set to this value
        """
        self.val = val


class ConstantMetric(BaseMetric):
    """Metric to track one fixed value

    This is generally used for logging strings

    Args:
        BaseMetric: Base metric class
    """

    def __init__(self, name: str, val: ValueType):
        self.name = name
        self.val = val

    def reset(self) -> None:
        """This function does not do anything"""
        return None

    def update(self, val: Optional[ValueType] = None) -> None:
        """This function does not do anything

        Args:
            val (Any): This value is ignored
        """
        return None


class ComparisonMetric(BaseMetric):
    """Metric to track the min/max value

    This is generally used for logging best accuracy, least loss, etc.

    Args:
        BaseMetric: Base metric class
    """

    def __init__(
        self, name: str, default_val: ValueType, comparison_op: ComparisonOpType
    ):
        """Metric to track the min/max value

        This is generally used for logging best accuracy, least loss, etc.

        Args:
            name (str): Name of the metric
            default_val (ValueType): Default value to initialise the metric
            comparison_op (ComparisonOpType): Operator to compare the current
                value with the incoming value.
                If comparison_op(current_val, new_val) is true, we update
                the current value.
        """
        self.name = name
        self._default_val = default_val
        self.comparison_op = comparison_op
        self.val = default_val

    def reset(self) -> None:
        """Reset the metric to the default value"""
        self.val = self._default_val

    def update(self, val: ValueType) -> None:
        """Use the comparison operator to decide which value to keep

        If the output of self.comparison_op(val, self)

        Args:
            val (ValueType): Value to compare the current value with.
                If comparison_op(current_val, new_val) is true, we update
                the current value.
        """
        if self.comparison_op(self.val, val):
            self.val = val


class MaxMetric(ComparisonMetric):
    """Metric to track the max value

    This is generally used for logging best accuracy, etc.

    Args:
        ComparisonMetric: Comparison metric class
    """

    def __init__(self, name: str):
        """Metric to track the max value

        This is generally used for logging best accuracy, etc.

        Args:
            name (str): Name of the metric
        """
        super().__init__(
            name=name, default_val=float("-inf"), comparison_op=operator.lt
        )


class MinMetric(ComparisonMetric):
    """Metric to track the min value

    This is generally used for logging least loss, etc.

    Args:
        ComparisonMetric: Comparison metric class
    """

    def __init__(self, name: str):
        """Metric to track the min value

        This is generally used for logging least loss, etc.

        Args:
            name (str): Name of the metric
        """
        super().__init__(name=name, default_val=float("inf"), comparison_op=operator.gt)


class AverageMetric(BaseMetric):
    """Metric to track the average value

    This is generally used for logging strings

    Args:
        BaseMetric: Base metric class
    """

    def __init__(self, name: str):
        self.name = name
        self.val: float
        self.avg: float
        self.sum: float
        self.count: float
        self.reset()

    def reset(self) -> None:
        self.val = 0.0
        self.avg = 0.0
        self.sum = 0.0
        self.count = 0.0

    def update(self, val: NumType, n: int = 1) -> None:
        """Update the metric using the current average value and the
            number of samples used to compute the average value

        Args:
            val (NumType): current average value
            n (int, optional): Number of samples used to compute the
                average. Defaults to 1
        """
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def get_val(self) -> float:
        """Get the current average value

        """
        return self.avg


class SumMetric(AverageMetric):
    """Metric to track the sum value

    Args:
        BaseMetric: Base metric class
    """

    def __init__(self, name: str):
        super().__init__(name)

    def get_val(self) -> float:
        """Get the current sum value

        """
        return self.sum


class MetricDict:
    """Class that wraps over a collection of metrics

    """

    def __init__(self, metric_list: Iterable[BaseMetric]):
        """Class that wraps over a collection of metrics

        Args:
            metric_list (Iterable[BaseMetric]): list of metrics to wrap
                over
        """
        self._metrics_dict = {metric.name: metric for metric in metric_list}

    def reset(self) -> None:
        """Reset all the metrics to default values
        """
        for key in self._metrics_dict:
            self._metrics_dict[key].reset()

    def update(self, metrics_dict: Union[LogType, "MetricDict"]) -> None:
        """Update all the metrics using the current values

        Args:
            metrics_dict (Union[LogType, MetricDict]): Current value of metrics
        """
        if isinstance(metrics_dict, MetricDict):
            metrics_dict = metrics_dict.to_dict()
        for key, val in metrics_dict.items():
            if key in self._metrics_dict:
                if isinstance(val, (str, float, int)):
                    self._metrics_dict[key].update(val)
                else:
                    self._metrics_dict[key].update(*val)

    def __str__(self) -> str:
        return "\n".join([repr(val) for key, val in self._metrics_dict.items()])

    def to_dict(self) -> LogType:
        """Method to convert the metrics into a dictionary that can be
            logged using the `LogBook`

        Returns:
            LogType: Metric data in as a dictionary
        """
        return {key: val.get_val() for key, val in self._metrics_dict.items()}



"""Implementation of Parser to parse the logs"""

import json
from typing import Callable, Dict, List, Optional

import pandas as pd

from ml_logger.parser import parser as base_parser
from ml_logger.parser import utils as parser_utils
from ml_logger.types import LogType, MetricType


def fn_to_group_metrics(metrics: List[MetricType]) -> Dict[str, List[MetricType]]:
    """Function to group a list of metrics into a dictionary of (key,
    list of grouped metrics)

    Args:
        metrics (List[MetricType]): List of metrics to group

    Returns:
        Dict[str, List[MetricType]]: Dictionary of (key,
            list of grouped metrics)
    """
    return {"all": metrics}


def fn_to_aggregate_metrics(metrics: List[MetricType]) -> List[MetricType]:
    """Function to aggregate a list of metrics

    Args:
        metrics (List[MetricType]): List of metrics to aggregate

    Returns:
        List[MetricType]: List of aggregated metrics
    """
    return metrics


class Parser(base_parser.Parser):
    """Class to parse the metrics in the log files
    """

    def __init__(
        self,
        log_transformer: Callable[
            [LogType], LogType
        ] = parser_utils.identity_log_transformer,
        error_handler: Callable[
            [str, json.decoder.JSONDecodeError], Optional[LogType]
        ] = parser_utils.silent_error_handler,
    ):
        """Class to parse the log files

        Args:
            log_transformer (Callable[[LogType], LogType], optional):
                Function to transform the logs after reading them from the
                filesystem. Defaults to parser_utils.identity_log_transformer.
            error_handler (Callable[[str, json.decoder.JSONDecodeError],
                Optional[LogType]], optional): Function to handle the
                error when the parser reads an invalid json string.
                Defaults to parser_utils.silent_error_handler.
        """

        super().__init__(
            log_transformer=log_transformer, error_handler=error_handler,
        )
        self.log_type = "metric"

    def get_metrics_as_df(
        self,
        file_path: str,
        fn_to_group_metrics: Callable[
            [List[LogType]], Dict[str, List[LogType]]
        ] = fn_to_group_metrics,
        fn_to_aggregate_metrics: Callable[
            [List[LogType]], List[LogType]
        ] = fn_to_aggregate_metrics,
    ) -> Dict[str, pd.DataFrame]:
        """Create a dict of (metric_name, dataframe)

        Method that:
        (i) reads the requested metrics (for each specified mode),
        (ii) groups metrics
        (iii) merge all the metrics in a group,
        (iii) converts the merged metrics into dataframes and returns a list of dataframes

        Args:
            file_path (str): Log file to read from
            fn_to_group_metrics (Callable[[List[LogType]], Dict[str, List[LogType]]], optional):
                Function to group a list of metrics into a dictionary of
                (key, list of grouped metrics). Defaults to fn_to_group_metrics.
            fn_to_aggregate_metrics (Callable[[List[LogType]], List[LogType]], optional):
                Function to aggregate a list of metrics. Defaults to fn_to_aggregate_metrics.

        Returns:
            Dict[str, pd.DataFrame]: [description]

        """
        metric_logs = list(self.get_logs(file_path=file_path))
        grouped_metrics: Dict[str, List[LogType]] = fn_to_group_metrics(metric_logs)
        merged_metrics = {
            key: fn_to_aggregate_metrics(metrics)
            for key, metrics in grouped_metrics.items()
        }

        metric_dfs = {
            key: pd.DataFrame.from_dict(
                data=parser_utils.map_list_of_dicts_to_dict_of_lists(metrics)
            )
            for key, metrics in merged_metrics.items()
        }
        return metric_dfs



"""Implementation of Parser to parse the logs"""

import glob
import json
from typing import Callable, Optional

from tinydb import TinyDB

from ml_logger.parser import parser as base_parser
from ml_logger.parser import utils as parser_utils
from ml_logger.types import ConfigType, LogType


class Parser(base_parser.Parser):
    """Class to parse the config in the log files
    """

    def __init__(
        self,
        log_transformer: Callable[
            [LogType], LogType
        ] = parser_utils.identity_log_transformer,
        error_handler: Callable[
            [str, json.decoder.JSONDecodeError], Optional[LogType]
        ] = parser_utils.silent_error_handler,
    ):
        """Class to parse the log files

        Args:
            log_transformer (Callable[[LogType], LogType], optional):
                Function to transform the logs after reading them from the
                filesystem. Defaults to parser_utils.identity_log_transformer.
            error_handler (Callable[[str, json.decoder.JSONDecodeError],
                Optional[LogType]], optional): Function to handle the
                error when the parser reads an invalid json string.
                Defaults to parser_utils.silent_error_handler.
        """

        super().__init__(
            log_transformer=log_transformer, error_handler=error_handler,
        )
        self.log_type = "config"

    def get_one_config(self, file_path: str) -> Optional[ConfigType]:
        """Method to get the config from the log file.

        The different between `get_config` and `get_logs` function is that
        `get_logs` returns an iterator over all the configs that are found
        in a log file while `get_config` only returns the last config. The same
        config file may be written multiple times in a log file. Hence
        `get_config` should be the default method for getting the config

        Args:
            file_path (str): Log file to read from

        Returns:
            ConfigType: Config object
        """

        configs = list(self.get_logs(file_path=file_path))
        if configs:
            return configs[-1]
        return None

    def load_configs_from_path(
        self, path_pattern: str, path_to_save: str = "config.json"
    ) -> TinyDB:
        """Method to glob the given path pattern and load config from
        all the matching paths

        Args:
            path_pattern (str): path pattern to glob
            path_to_save (str): filesystem path where the loaded configs
                are saved. Defaults to config.json.

        Returns:
            TinyDB: TinyDB database instance over the collection of configs
        """
        db = TinyDB(path_to_save)
        paths = glob.glob(path_pattern)
        for file_path in paths:
            config = self.get_one_config(file_path=file_path)
            if config is not None:
                config["file_path"] = file_path
                db.insert(config)
        return db



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mllogger",
    version="0.5rc1",
    author="Shagun Sodhani",
    author_email="sshagunsodhani@gmail.com",
    description="Logging Utility for ML Experiments",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    # Install the basic setup (without wandb, tensorboardX and mlflow) with
    install_requires=open("requirements/filesystem.txt").read().splitlines(),
    url="https://github.com/shagunsodhani/ml-logger",
    packages=setuptools.find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests", "docs", "docsrc"]
    ),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    extras_require={
        # Install development dependencies with
        # pip install -e .[dev]
        "dev": open("requirements/dev.txt").read().splitlines(),
        # Install the complete setup (wandb, mlflow and tensorboardX)
        # pip install .[all]
        "all": open("requirements/all.txt").read().splitlines(),
    },
)



import json
from typing import Any, Dict, List, Optional, Tuple

from ml_logger.types import LogType, ValueType


def map_list_of_dicts_to_dict_of_lists(
    list_of_dicts: List[Dict[str, ValueType]]
) -> Dict[str, List[Optional[ValueType]]]:
    """Map a list of dictionary to a dictionary of lists

    Example input: [
        {"a": 1, "b": 2},
        {"b": 3, "c": 4},
    ]

    Example output: {
        "a": [1],
        "b": [2, 3],
        "c": [4]
    }

    Args:
        list_of_dicts (List[Dict[str, ValueType]]): List of dictionaries

    Returns:
        Dict[str, List[Optional[ValueType]]]: Dictionary of lists
    """
    if not list_of_dicts:
        return {}
    keys = list_of_dicts[0].keys()
    dict_of_lists = {}
    for key in keys:
        dict_of_lists[key] = [_dict.get(key, None) for _dict in list_of_dicts]
    return dict_of_lists


def identity_log_transformer(log: LogType) -> LogType:
    """Function to transform the log after parsing

    Args:
        log (LogType): log to transform

    Returns:
        LogType: transformed log
    """
    return log


def verbose_error_handler(
    log_line: str, error: json.decoder.JSONDecodeError
) -> Optional[LogType]:
    """Function to print the error on the console, when the `log_line` is
        not a valid json string

    Args:
        log_line (str): Parsing this line triggered the error
        error (json.decoder.JSONDecodeError): The error object

    Returns:
        Optional[LogType]: None. Print the error on the console
    """
    print(f"Could not parse: {log_line} because of error: {error}")
    return None


def silent_error_handler(
    log_line: str, error: json.decoder.JSONDecodeError
) -> Optional[LogType]:
    """Function to silently ignore the error, when the `log_line` is
        not a valid json string

    Args:
        log_line (str): Parsing this line triggered the error
        error (json.decoder.JSONDecodeError): The error object

    Returns:
        Optional[LogType]: None. Nothing is done
    """
    return None


def flatten_log(d: LogType, parent_key: str = "", sep: str = "#") -> LogType:
    """Flatten a log using a separator

    Taken from https://stackoverflow.com/a/6027615/1353861

    Args:
        d (LogType): [description]
        parent_key (str, optional): [description]. Defaults to "".
        sep (str, optional): [description]. Defaults to "#".

    Returns:
        LogType: [description]
    """
    items: List[Tuple[str, Any]] = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_log(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def compare_logs(
    first_log: LogType, second_log: LogType, verbose: bool = False
) -> Tuple[List[str], List[str], List[str]]:
    """Compare two logs and return list of keys that are either missing or
        have different valus in the two logs.

    Args:
        first_log (LogType): First Log
        second_log (LogType): Second Log
        verbose (bool): Defaults to False

    Returns:
        Tuple[List[str], List[str], List[str]]: Tuple of [
            list of keys with different values,
            list of keys with values missing in first log,
            list of keys with values missing in the second log
        ]
    """
    first_log = flatten_log(first_log)
    second_log = flatten_log(second_log)
    first_keys = set(first_log.keys())
    second_keys = set(second_log.keys())
    keys = first_keys.union(second_keys)
    keys_with_diff_values = []
    keys_with_missing_value_in_first_log = []
    keys_with_missing_value_in_second_log = []
    for key in keys:
        if key not in first_log:
            keys_with_missing_value_in_first_log.append(key)
            if verbose:
                print(f"first_log[{key}]: ???, second_log[{key}]: {second_log[key]}")
        elif key not in second_log:
            keys_with_missing_value_in_second_log.append(key)
            print(f"first_log[{key}]: {first_log[key]},  second_log[{key}]: ???")
        else:
            if first_log[key] != second_log[key]:
                keys_with_diff_values.append(key)
                print(
                    f"first_log[{key}]: {first_log[key]},  second_log[{key}]: {second_log[key]}"
                )
    return (
        keys_with_diff_values,
        keys_with_missing_value_in_first_log,
        keys_with_missing_value_in_second_log,
    )
